#!/usr/bin/python3
# coding: utf-8
import os
import sys
from setuptools import setup
from setuptools.command.install import install

VERSION = "3.2.1"

with open("README.md", "r") as fh:
    long_description = fh.read()


class VerifyVersionCommand(install):
    description = "verify that the git tag matches our version"

    def run(self):
        tag = os.getenv("CIRCLE_TAG")
        if tag != VERSION:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)


setup(
    name="pycti",
    version=VERSION,
    description="Python API client for OpenCTI.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="OpenCTI",
    author_email="contact@opencti.io",
    maintainer="OpenCTI",
    url="https://github.com/OpenCTI-Platform/client-python",
    license="Apache",
    packages=["pycti", "pycti.api", "pycti.connector", "pycti.entities", "pycti.utils"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Natural Language :: French",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Security",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    include_package_data=True,
    install_requires=[
        "requests==2.23.0",
        "PyYAML==5.3.1",
        "python-dateutil==2.8.1",
        "datefinder==0.7.0",
        "stix2==1.4.0",
        "stix2-validator==1.1.2",
        "pytz==2019.3",
        "pika==1.1.0",
        "python-magic==0.4.15;sys.platform=='linux' or sys.platform=='darwin'",
        "python-magic-bin==0.4.14;sys.platform=='win32'",
    ],
    cmdclass={"verify": VerifyVersionCommand},
    extras_require={"dev": ["black", "wheel"]},  # Optional
)



# coding: utf-8

import io
import magic
import requests
import urllib3
import json
import logging

from pycti.api.opencti_api_connector import OpenCTIApiConnector
from pycti.api.opencti_api_job import OpenCTIApiJob
from pycti.utils.constants import ObservableTypes
from pycti.utils.opencti_stix2 import OpenCTIStix2

from pycti.entities.opencti_tag import Tag
from pycti.entities.opencti_marking_definition import MarkingDefinition
from pycti.entities.opencti_external_reference import ExternalReference
from pycti.entities.opencti_kill_chain_phase import KillChainPhase
from pycti.entities.opencti_stix_entity import StixEntity
from pycti.entities.opencti_stix_domain_entity import StixDomainEntity
from pycti.entities.opencti_stix_observable import StixObservable
from pycti.entities.opencti_stix_relation import StixRelation
from pycti.entities.opencti_stix_observable_relation import StixObservableRelation
from pycti.entities.opencti_identity import Identity
from pycti.entities.opencti_threat_actor import ThreatActor
from pycti.entities.opencti_intrusion_set import IntrusionSet
from pycti.entities.opencti_campaign import Campaign
from pycti.entities.opencti_incident import Incident
from pycti.entities.opencti_malware import Malware
from pycti.entities.opencti_tool import Tool
from pycti.entities.opencti_vulnerability import Vulnerability
from pycti.entities.opencti_attack_pattern import AttackPattern
from pycti.entities.opencti_course_of_action import CourseOfAction
from pycti.entities.opencti_report import Report
from pycti.entities.opencti_note import Note
from pycti.entities.opencti_opinion import Opinion
from pycti.entities.opencti_indicator import Indicator

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class File:
    def __init__(self, name, data, mime="text/plain"):
        self.name = name
        self.data = data
        self.mime = mime


class OpenCTIApiClient:
    """
        Python API for OpenCTI
        :param url: OpenCTI URL
        :param token: The API key
    """

    def __init__(self, url, token, log_level="info", ssl_verify=False):
        # Check configuration
        self.ssl_verify = ssl_verify
        if url is None or len(token) == 0:
            raise ValueError("Url configuration must be configured")
        if token is None or len(token) == 0 or token == "ChangeMe":
            raise ValueError(
                "Token configuration must be the same as APP__ADMIN__TOKEN"
            )

        # Configure logger
        self.log_level = log_level
        numeric_level = getattr(logging, self.log_level.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError("Invalid log level: " + self.log_level)
        logging.basicConfig(level=numeric_level)

        # Define API
        self.api_token = token
        self.api_url = url + "/graphql"
        self.request_headers = {"Authorization": "Bearer " + token}

        # Define the dependencies
        self.job = OpenCTIApiJob(self)
        self.connector = OpenCTIApiConnector(self)
        self.stix2 = OpenCTIStix2(self)

        # Define the entities
        self.tag = Tag(self)
        self.marking_definition = MarkingDefinition(self)
        self.external_reference = ExternalReference(self)
        self.kill_chain_phase = KillChainPhase(self)
        self.stix_entity = StixEntity(self)
        self.stix_domain_entity = StixDomainEntity(self, File)
        self.stix_observable = StixObservable(self)
        self.stix_relation = StixRelation(self)
        self.stix_observable_relation = StixObservableRelation(self)
        self.identity = Identity(self)
        self.threat_actor = ThreatActor(self)
        self.intrusion_set = IntrusionSet(self)
        self.campaign = Campaign(self)
        self.incident = Incident(self)
        self.malware = Malware(self)
        self.tool = Tool(self)
        self.vulnerability = Vulnerability(self)
        self.attack_pattern = AttackPattern(self)
        self.course_of_action = CourseOfAction(self)
        self.report = Report(self)
        self.note = Note(self)
        self.opinion = Opinion(self)
        self.indicator = Indicator(self)

        # Check if openCTI is available
        if not self.health_check():
            raise ValueError(
                "OpenCTI API is not reachable. Waiting for OpenCTI API to start or check your configuration..."
            )

    def get_token(self):
        return self.api_token

    def set_token(self, token):
        self.request_headers = {"Authorization": "Bearer " + token}

    def query(self, query, variables={}):
        query_var = {}
        files_vars = []
        # Implementation of spec https://github.com/jaydenseric/graphql-multipart-request-spec
        # Support for single or multiple upload
        # Batching or mixed upload or not supported
        var_keys = variables.keys()
        for key in var_keys:
            val = variables[key]
            is_file = type(val) is File
            is_files = (
                isinstance(val, list)
                and len(val) > 0
                and all(map(lambda x: isinstance(x, File), val))
            )
            if is_file or is_files:
                files_vars.append({"key": key, "file": val, "multiple": is_files})
                query_var[key] = None if is_file else [None] * len(val)
            else:
                query_var[key] = val

        # If yes, transform variable (file to null) and create multipart query
        if len(files_vars) > 0:
            multipart_data = {
                "operations": json.dumps({"query": query, "variables": query_var})
            }
            # Build the multipart map
            map_index = 0
            file_vars = {}
            for file_var_item in files_vars:
                is_multiple_files = file_var_item["multiple"]
                var_name = "variables." + file_var_item["key"]
                if is_multiple_files:
                    # [(var_name + "." + i)] if is_multiple_files else
                    for _ in file_var_item["file"]:
                        file_vars[str(map_index)] = [(var_name + "." + str(map_index))]
                        map_index += 1
                else:
                    file_vars[str(map_index)] = [var_name]
                    map_index += 1
            multipart_data["map"] = json.dumps(file_vars)
            # Add the files
            file_index = 0
            multipart_files = []
            for file_var_item in files_vars:
                files = file_var_item["file"]
                is_multiple_files = file_var_item["multiple"]
                if is_multiple_files:
                    for file in files:
                        if isinstance(file.data, str):
                            file_multi = (
                                str(file_index),
                                (file.name, io.BytesIO(file.data.encode()), file.mime,),
                            )
                        else:
                            file_multi = (
                                str(file_index),
                                (file.name, file.data, file.mime),
                            )
                        multipart_files.append(file_multi)
                        file_index += 1
                else:
                    if isinstance(files.data, str):
                        file_multi = (
                            str(file_index),
                            (files.name, io.BytesIO(files.data.encode()), files.mime),
                        )
                    else:
                        file_multi = (
                            str(file_index),
                            (files.name, files.data, files.mime),
                        )
                    multipart_files.append(file_multi)
                    file_index += 1
            # Send the multipart request
            r = requests.post(
                self.api_url,
                data=multipart_data,
                files=multipart_files,
                headers=self.request_headers,
                verify=self.ssl_verify,
            )
        # If no
        else:
            r = requests.post(
                self.api_url,
                json={"query": query, "variables": variables},
                headers=self.request_headers,
                verify=self.ssl_verify,
            )
        # Build response
        if r.status_code == requests.codes.ok:
            result = r.json()
            if "errors" in result:
                logging.error(result["errors"][0]["message"])
            else:
                return result
        else:
            logging.info(r.text)

    def fetch_opencti_file(self, fetch_uri, binary=False):
        r = requests.get(fetch_uri, headers=self.request_headers)
        if binary:
            return r.content
        return r.text

    def log(self, level, message):
        if level == "debug":
            logging.debug(message)
        elif level == "info":
            logging.info(message)
        elif level == "warning":
            logging.warn(message)
        elif level == "error":
            logging.error(message)

    def health_check(self):
        try:
            test = self.threat_actor.list(first=1)
            if test is not None:
                return True
        except:
            return False
        return False

    def get_logs_worker_config(self):
        logging.info("Getting logs worker config...")
        query = """
            query LogsWorkerConfig {
                logsWorkerConfig {
                    elasticsearch_url
                    elasticsearch_index
                    rabbitmq_url
                }
            }
        """
        result = self.query(query)
        return result["data"]["logsWorkerConfig"]

    def not_empty(self, value):
        if value is not None:
            if isinstance(value, str):
                if len(value) > 0:
                    return True
                else:
                    return False
            if isinstance(value, list):
                is_not_empty = False
                for v in value:
                    if len(v) > 0:
                        is_not_empty = True
                return is_not_empty
            if isinstance(value, int):
                return True
            else:
                return False
        else:
            return False

    def process_multiple(self, data):
        result = []
        if data is None:
            return result
        for edge in (
            data["edges"] if "edges" in data and data["edges"] is not None else []
        ):
            row = edge["node"]
            # Handle remote relation ID
            if (
                "relation" in edge
                and edge["relation"] is not None
                and "id" in edge["relation"]
            ):
                row["remote_relation_id"] = edge["relation"]["id"]
            result.append(self.process_multiple_fields(row))
        return result

    def process_multiple_ids(self, data):
        result = []
        if data is None:
            return result
        if isinstance(data, list):
            for d in data:
                if isinstance(d, dict) and "id" in d:
                    result.append(d["id"])
        return result

    def process_multiple_fields(self, data):
        if data is None:
            return data
        if (
            "createdByRef" in data
            and data["createdByRef"] is not None
            and "node" in data["createdByRef"]
        ):
            row = data["createdByRef"]["node"]
            # Handle remote relation ID
            if "relation" in data["createdByRef"]:
                row["remote_relation_id"] = data["createdByRef"]["relation"]["id"]
            data["createdByRef"] = row
        else:
            data["createdByRef"] = None
        if "markingDefinitions" in data:
            data["markingDefinitions"] = self.process_multiple(
                data["markingDefinitions"]
            )
            data["markingDefinitionsIds"] = self.process_multiple_ids(
                data["markingDefinitions"]
            )
        if "tags" in data:
            data["tags"] = self.process_multiple(data["tags"])
            data["tagsIds"] = self.process_multiple_ids(data["tags"])
        if "reports" in data:
            data["reports"] = self.process_multiple(data["reports"])
            data["reportsIds"] = self.process_multiple_ids(data["reports"])
        if "notes" in data:
            data["notes"] = self.process_multiple(data["notes"])
            data["notesIds"] = self.process_multiple_ids(data["notes"])
        if "opinions" in data:
            data["opinions"] = self.process_multiple(data["opinions"])
            data["opinionsIds"] = self.process_multiple_ids(data["opinions"])
        if "killChainPhases" in data:
            data["killChainPhases"] = self.process_multiple(data["killChainPhases"])
            data["killChainPhasesIds"] = self.process_multiple_ids(
                data["killChainPhases"]
            )
        if "externalReferences" in data:
            data["externalReferences"] = self.process_multiple(
                data["externalReferences"]
            )
            data["externalReferencesIds"] = self.process_multiple_ids(
                data["externalReferences"]
            )
        if "objectRefs" in data:
            data["objectRefs"] = self.process_multiple(data["objectRefs"])
            data["objectRefsIds"] = self.process_multiple_ids(data["objectRefs"])
        if "observableRefs" in data:
            data["observableRefs"] = self.process_multiple(data["observableRefs"])
            data["observableRefsIds"] = self.process_multiple_ids(
                data["observableRefs"]
            )
        if "relationRefs" in data:
            data["relationRefs"] = self.process_multiple(data["relationRefs"])
            data["relationRefsIds"] = self.process_multiple_ids(data["relationRefs"])
        if "stixRelations" in data:
            data["stixRelations"] = self.process_multiple(data["stixRelations"])
            data["stixRelationsIds"] = self.process_multiple_ids(data["stixRelations"])
        if "indicators" in data:
            data["indicators"] = self.process_multiple(data["indicators"])
            data["indicatorsIds"] = self.process_multiple_ids(data["indicators"])
        if "importFiles" in data:
            data["importFiles"] = self.process_multiple(data["importFiles"])
            data["importFilesIds"] = self.process_multiple_ids(data["importFiles"])
        return data

    def upload_file(self, **kwargs):
        file_name = kwargs.get("file_name", None)
        data = kwargs.get("data", None)
        mime_type = kwargs.get("mime_type", "text/plain")
        if file_name is not None:
            self.log("info", "Uploading a file.")
            query = """
                mutation UploadImport($file: Upload!) {
                    uploadImport(file: $file) {
                        id
                        name
                    }
                }
             """
            if data is None:
                data = open(file_name, "rb")
                mime_type = magic.from_file(file_name, mime=True)

            return self.query(query, {"file": (File(file_name, data, mime_type))})
        else:
            self.log(
                "error", "[upload] Missing parameters: file_name or data",
            )
            return None

    # TODO Move to StixObservable
    def update_stix_observable_field(self, id, key, value):
        logging.info("Updating field " + key + " of " + id + "...")
        query = """
            mutation StixObservableEdit($id: ID!, $input: EditInput!) {
                stixObservableEdit(id: $id) {
                    fieldPatch(input: $input) {
                        id
                        observable_value
                        entity_type
                    }
                }
            }
        """
        self.query(query, {"id": id, "input": {"key": key, "value": value}})

    # TODO Move to ExternalReference
    def delete_external_reference(self, id):
        logging.info("Deleting + " + id + "...")
        query = """
             mutation ExternalReferenceEdit($id: ID!) {
                 externalReferenceEdit(id: $id) {
                     delete
                 }
             }
         """
        self.query(query, {"id": id})

    # TODO Move to Vulnerability
    def create_vulnerability_if_not_exists(
        self,
        name,
        description,
        alias=None,
        id=None,
        stix_id_key=None,
        created=None,
        modified=None,
        update=False,
    ):
        return self.vulnerability.create(
            name=name,
            description=description,
            alias=alias,
            id=id,
            stix_id_key=stix_id_key,
            created=created,
            modified=modified,
            update=update,
        )

    def resolve_role(self, relation_type, from_type, to_type):
        if from_type == "stix-relation":
            from_type = "stix_relation"
        if to_type == "stix-relation":
            to_type = "stix_relation"
        if relation_type == "related-to":
            return {"from_role": "relate_from", "to_role": "relate_to"}

        relation_type = relation_type.lower()
        from_type = from_type.lower()
        from_type = (
            "observable"
            if (
                (
                    ObservableTypes.has_value(from_type)
                    and (
                        relation_type == "localization" or relation_type == "gathering"
                    )
                )
                or from_type == "stix-observable"
            )
            else from_type
        )
        to_type = to_type.lower()
        mapping = {
            "uses": {
                "threat-actor": {
                    "malware": {"from_role": "user", "to_role": "usage"},
                    "tool": {"from_role": "user", "to_role": "usage"},
                    "attack-pattern": {"from_role": "user", "to_role": "usage"},
                },
                "intrusion-set": {
                    "malware": {"from_role": "user", "to_role": "usage"},
                    "tool": {"from_role": "user", "to_role": "usage"},
                    "attack-pattern": {"from_role": "user", "to_role": "usage"},
                },
                "campaign": {
                    "malware": {"from_role": "user", "to_role": "usage"},
                    "tool": {"from_role": "user", "to_role": "usage"},
                    "attack-pattern": {"from_role": "user", "to_role": "usage"},
                },
                "incident": {
                    "malware": {"from_role": "user", "to_role": "usage"},
                    "tool": {"from_role": "user", "to_role": "usage"},
                    "attack-pattern": {"from_role": "user", "to_role": "usage"},
                },
                "malware": {
                    "tool": {"from_role": "user", "to_role": "usage"},
                    "attack-pattern": {"from_role": "user", "to_role": "usage"},
                },
                "tool": {"attack-pattern": {"from_role": "user", "to_role": "usage"}},
            },
            "variant-of": {
                "malware": {
                    "malware": {"from_role": "original", "to_role": "variation"},
                },
                "tool": {"tool": {"from_role": "original", "to_role": "variation"},},
            },
            "targets": {
                "threat-actor": {
                    "identity": {"from_role": "source", "to_role": "target"},
                    "sector": {"from_role": "source", "to_role": "target"},
                    "region": {"from_role": "source", "to_role": "target"},
                    "country": {"from_role": "source", "to_role": "target"},
                    "city": {"from_role": "source", "to_role": "target"},
                    "organization": {"from_role": "source", "to_role": "target"},
                    "user": {"from_role": "source", "to_role": "target"},
                    "vulnerability": {"from_role": "source", "to_role": "target"},
                },
                "intrusion-set": {
                    "identity": {"from_role": "source", "to_role": "target"},
                    "sector": {"from_role": "source", "to_role": "target"},
                    "region": {"from_role": "source", "to_role": "target"},
                    "country": {"from_role": "source", "to_role": "target"},
                    "city": {"from_role": "source", "to_role": "target"},
                    "organization": {"from_role": "source", "to_role": "target"},
                    "user": {"from_role": "source", "to_role": "target"},
                    "vulnerability": {"from_role": "source", "to_role": "target"},
                },
                "campaign": {
                    "identity": {"from_role": "source", "to_role": "target"},
                    "sector": {"from_role": "source", "to_role": "target"},
                    "region": {"from_role": "source", "to_role": "target"},
                    "country": {"from_role": "source", "to_role": "target"},
                    "city": {"from_role": "source", "to_role": "target"},
                    "organization": {"from_role": "source", "to_role": "target"},
                    "user": {"from_role": "source", "to_role": "target"},
                    "vulnerability": {"from_role": "source", "to_role": "target"},
                },
                "incident": {
                    "identity": {"from_role": "source", "to_role": "target"},
                    "sector": {"from_role": "source", "to_role": "target"},
                    "region": {"from_role": "source", "to_role": "target"},
                    "country": {"from_role": "source", "to_role": "target"},
                    "city": {"from_role": "source", "to_role": "target"},
                    "organization": {"from_role": "source", "to_role": "target"},
                    "user": {"from_role": "source", "to_role": "target"},
                    "vulnerability": {"from_role": "source", "to_role": "target"},
                },
                "malware": {
                    "identity": {"from_role": "source", "to_role": "target"},
                    "sector": {"from_role": "source", "to_role": "target"},
                    "region": {"from_role": "source", "to_role": "target"},
                    "country": {"from_role": "source", "to_role": "target"},
                    "city": {"from_role": "source", "to_role": "target"},
                    "organization": {"from_role": "source", "to_role": "target"},
                    "user": {"from_role": "source", "to_role": "target"},
                    "vulnerability": {"from_role": "source", "to_role": "target"},
                },
                "attack-pattern": {
                    "vulnerability": {"from_role": "source", "to_role": "target"},
                },
            },
            "attributed-to": {
                "threat-actor": {
                    "identity": {"from_role": "attribution", "to_role": "origin"},
                    "organization": {"from_role": "attribution", "to_role": "origin"},
                    "user": {"from_role": "attribution", "to_role": "origin"},
                },
                "intrusion-set": {
                    "identity": {"from_role": "attribution", "to_role": "origin"},
                    "threat-actor": {"from_role": "attribution", "to_role": "origin"},
                },
                "campaign": {
                    "identity": {"from_role": "attribution", "to_role": "origin"},
                    "threat-actor": {"from_role": "attribution", "to_role": "origin"},
                    "intrusion-set": {"from_role": "attribution", "to_role": "origin"},
                },
                "incident": {
                    "identity": {"from_role": "attribution", "to_role": "origin"},
                    "threat-actor": {"from_role": "attribution", "to_role": "origin"},
                    "intrusion-set": {"from_role": "attribution", "to_role": "origin"},
                    "campaign": {"from_role": "attribution", "to_role": "origin"},
                },
                "malware": {
                    "identity": {"from_role": "attribution", "to_role": "origin"},
                    "threat-actor": {"from_role": "attribution", "to_role": "origin"},
                },
            },
            "mitigates": {
                "course-of-action": {
                    "attack-pattern": {"from_role": "mitigation", "to_role": "problem"}
                }
            },
            "localization": {
                "threat-actor": {
                    "region": {"from_role": "localized", "to_role": "location"},
                    "country": {"from_role": "localized", "to_role": "location"},
                    "city": {"from_role": "localized", "to_role": "location"},
                },
                "observable": {
                    "region": {"from_role": "localized", "to_role": "location"},
                    "country": {"from_role": "localized", "to_role": "location"},
                    "city": {"from_role": "localized", "to_role": "location"},
                },
                "stix_relation": {
                    "region": {"from_role": "localized", "to_role": "location"},
                    "country": {"from_role": "localized", "to_role": "location"},
                    "city": {"from_role": "localized", "to_role": "location"},
                },
                "region": {"region": {"from_role": "localized", "to_role": "location"}},
                "country": {
                    "region": {"from_role": "localized", "to_role": "location"}
                },
                "city": {"country": {"from_role": "localized", "to_role": "location"}},
                "organization": {
                    "region": {"from_role": "localized", "to_role": "location"},
                    "country": {"from_role": "localized", "to_role": "location"},
                    "city": {"from_role": "localized", "to_role": "location"},
                },
            },
            "indicates": {
                "indicator": {
                    "threat-actor": {
                        "from_role": "indicator",
                        "to_role": "characterize",
                    },
                    "intrusion-set": {
                        "from_role": "indicator",
                        "to_role": "characterize",
                    },
                    "campaign": {"from_role": "indicator", "to_role": "characterize"},
                    "malware": {"from_role": "indicator", "to_role": "characterize"},
                    "tool": {"from_role": "indicator", "to_role": "characterize"},
                    "stix_relation": {
                        "from_role": "indicator",
                        "to_role": "characterize",
                    },
                }
            },
            "gathering": {
                "threat-actor": {
                    "organization": {"from_role": "part_of", "to_role": "gather"},
                },
                "sector": {
                    "sector": {"from_role": "part_of", "to_role": "gather"},
                    "organization": {"from_role": "part_of", "to_role": "gather"},
                },
                "organization": {
                    "sector": {"from_role": "part_of", "to_role": "gather"},
                    "organization": {"from_role": "part_of", "to_role": "gather"},
                },
                "user": {
                    "organization": {"from_role": "part_of", "to_role": "gather"},
                },
                "observable": {
                    "organization": {"from_role": "part_of", "to_role": "gather"},
                    "user": {"from_role": "part_of", "to_role": "gather"},
                },
            },
            "drops": {
                "malware": {
                    "malware": {"from_role": "dropping", "to_role": "dropped"},
                    "tool": {"from_role": "dropping", "to_role": "dropped"},
                },
                "tool": {
                    "malware": {"from_role": "dropping", "to_role": "dropped"},
                    "tool": {"from_role": "dropping", "to_role": "dropped"},
                },
            },
            "belongs": {
                "ipv4-addr": {
                    "autonomous-system": {
                        "from_role": "belonging_to",
                        "to_role": "belonged_to",
                    }
                },
                "ipv6-addr": {
                    "autonomous-system": {
                        "from_role": "belonging_to",
                        "to_role": "belonged_to",
                    }
                },
            },
            "resolves": {
                "ipv4-addr": {
                    "domain": {"from_role": "resolving", "to_role": "resolved",}
                },
                "ipv6-addr": {
                    "domain": {"from_role": "resolving", "to_role": "resolved",}
                },
            },
            "corresponds": {
                "file-name": {
                    "file-md5": {
                        "from_role": "correspond_from",
                        "to_role": "correspond_to",
                    },
                    "file-sha1": {
                        "from_role": "correspond_from",
                        "to_role": "correspond_to",
                    },
                    "file-sha256": {
                        "from_role": "correspond_from",
                        "to_role": "correspond_to",
                    },
                },
                "file-md5": {
                    "file-name": {
                        "from_role": "correspond_from",
                        "to_role": "correspond_to",
                    },
                    "file-sha1": {
                        "from_role": "correspond_from",
                        "to_role": "correspond_to",
                    },
                    "file-sha256": {
                        "from_role": "correspond_from",
                        "to_role": "correspond_to",
                    },
                },
                "file-sha1": {
                    "file-name": {
                        "from_role": "correspond_from",
                        "to_role": "correspond_to",
                    },
                    "file-md5": {
                        "from_role": "correspond_from",
                        "to_role": "correspond_to",
                    },
                    "file-sha256": {
                        "from_role": "correspond_from",
                        "to_role": "correspond_to",
                    },
                },
                "file-sha256": {
                    "file-name": {
                        "from_role": "correspond_from",
                        "to_role": "correspond_to",
                    },
                    "file-md5": {
                        "from_role": "correspond_from",
                        "to_role": "correspond_to",
                    },
                    "file-sha1": {
                        "from_role": "correspond_from",
                        "to_role": "correspond_to",
                    },
                },
            },
        }
        if (
            relation_type in mapping
            and from_type in mapping[relation_type]
            and to_type in mapping[relation_type][from_type]
        ):
            return mapping[relation_type][from_type][to_type]
        else:
            return None



"""These are the custom STIX properties and observation types used internally by OpenCTI.

"""
from enum import Enum


class ObservableTypes(Enum):
    """These are the possible values for OpenCTI's observable types.
    Use in conjuction with the STIX custom property 'x_opencti_observable_type'.
    ref: https://github.com/OpenCTI-Platform/opencti/blob/8854c2576dc17da9da54e54b116779bd2131617c/opencti-front/src/private/components/report/ReportAddObservable.js
    NOTE: should this be a mapping between the stix2 SDO objects (i.e. stix2/v20/sdo.py)?
    """

    AUTONOMOUS_SYSTEM = "Autonomous-System"
    DOMAIN = "Domain"
    EMAIL_ADDR = "Email-Address"
    EMAIL_SUBJECT = "Email-Subject"
    DIRECTORY = "Directory"
    FILE_NAME = "File-Name"
    FILE_PATH = "File-Path"
    FILE_HASH_MD5 = "File-MD5"
    FILE_HASH_SHA1 = "File-SHA1"
    FILE_HASH_SHA256 = "File-SHA256"
    IPV4_ADDR = "IPv4-Addr"
    IPV6_ADDR = "IPv6-Addr"
    MAC_ADDR = "Mac-Addr"
    MUTEX = "Mutex"
    PDB_PATH = "PDB-Path"
    REGISTRY_KEY = "Registry-Key"
    REGISTRY_VALUE = "Registry-Key-Value"
    URL = "URL"
    WIN_SERVICE_NAME = "Windows-Service-Name"
    WIN_SERVICE_DISPLAY = "Windows-Service-Display-Name"
    WIN_SCHEDULED_TASK = "Windows-Scheduled-Task"
    X509_CERT_ISSUER = "X509-Certificate-Issuer"
    X509_CERT_SN = "X509-Certificate-Serial-Number"

    @classmethod
    def has_value(cls, value):
        lower_attr = list(map(lambda x: x.lower(), cls._value2member_map_))
        return value in lower_attr


class IdentityTypes(Enum):
    SECTOR = "Sector"
    REGION = "Region"
    COUNTRY = "Country"
    CITY = "City"
    ORGANIZATION = "Organization"
    USER = "User"

    @classmethod
    def has_value(cls, value):
        lower_attr = list(map(lambda x: x.lower(), cls._value2member_map_))
        return value in lower_attr


class StixObservableRelationTypes(Enum):
    LINKED = "linked"
    RESOLVES = "resolves"
    BELONGS = "belongs"
    CONTAINS = "contains"
    CORRESPONDS = "corresponds"

    @classmethod
    def has_value(cls, value):
        lower_attr = list(map(lambda x: x.lower(), cls._value2member_map_))
        return value in lower_attr


class CustomProperties:
    """These are the custom properies used by OpenCTI.

    """

    # internal id used by OpenCTI - this will be auto generated
    ID = "x_opencti_id"

    # List of files
    FILES = "x_opencti_files"

    # This should be set on all reports to one of the following values:
    #  "external"
    #  "internal"
    REPORT_CLASS = "x_opencti_report_class"

    # use with observed_data and indicators
    INDICATOR_PATTERN = "x_opencti_indicator_pattern"
    PATTERN_TYPE = "x_opencti_pattern_type"
    OBSERVABLE_TYPE = "x_opencti_observable_type"
    OBSERVABLE_VALUE = "x_opencti_observable_value"
    CREATE_OBSERVABLES = "x_opencti_observables_create"
    CREATE_INDICATOR = "x_opencti_indicator_create"

    # custom created and modified dates
    # use with STIX "kill chain" and "external reference" objects
    CREATED = "x_opencti_created"
    MODIFIED = "x_opencti_modified"

    # use with attack pattern
    EXTERNAL_ID = "x_opencti_external_id"

    # use with vulnerability
    BASE_SCORE = "x_opencti_base_score"
    BASE_SEVERITY = "x_opencti_base_severity"
    ATTACK_VECTOR = "x_opencti_attack_vector"
    INTEGRITY_IMPACT = "x_opencti_integrity_impact"
    AVAILABILITY_IMPACT = "x_opencti_availability_impact"

    # use with intrusion-set, campaign, relation
    FIRST_SEEN = "x_opencti_first_seen"
    LAST_SEEN = "x_opencti_last_seen"

    # use with marking definitions
    COLOR = "x_opencti_color"
    LEVEL = "x_opencti_level"  # should be an integer

    # use with kill chain
    PHASE_ORDER = "x_opencti_phase_order"

    # use with relation
    WEIGHT = "x_opencti_weight"
    SCORE = "x_opencti_score"
    ROLE_PLAYED = "x_opencti_role_played"
    EXPIRATION = "x_opencti_expiration"
    SOURCE_REF = "x_opencti_source_ref"
    TARGET_REF = "x_opencti_target_ref"
    IGNORE_DATES = "x_opencti_ignore_dates"

    # generic property - applies to most SDOs
    ALIASES = "x_opencti_aliases"

    # applies to STIX Identity
    ORG_CLASS = "x_opencti_organization_class"
    IDENTITY_TYPE = (
        "x_opencti_identity_type"  # this overrides the stix 'identity_class' property!
    )
    TAG_TYPE = "x_opencti_tags"

    # applies to STIX report
    OBJECT_STATUS = "x_opencti_object_status"
    SRC_CONF_LEVEL = "x_opencti_source_confidence_level"
    GRAPH_DATA = "x_opencti_graph_data"

    # applies to STIX note
    NAME = "x_opencti_name"



# coding: utf-8

import json
from pycti.utils.constants import CustomProperties
from pycti.utils.opencti_stix2 import SPEC_VERSION


class AttackPattern:
    def __init__(self, opencti):
        self.opencti = opencti
        self.properties = """
            id
            stix_id_key
            stix_label
            entity_type
            parent_types
            name
            alias
            description
            graph_data
            platform
            required_permission
            external_id
            created
            modified
            created_at
            updated_at
            killChainPhases {
                edges {
                    node {
                        id
                        entity_type
                        stix_id_key
                        kill_chain_name
                        phase_name
                        phase_order
                        created
                        modified
                    }
                    relation {
                        id
                    }
                }
            }
            createdByRef {
                node {
                    id
                    entity_type
                    stix_id_key
                    stix_label
                    name
                    alias
                    description
                    created
                    modified
                    ... on Organization {
                        organization_class
                    }
                }
                relation {
                    id
                }
            }            
            markingDefinitions {
                edges {
                    node {
                        id
                        entity_type
                        stix_id_key
                        definition_type
                        definition
                        level
                        color
                        created
                        modified
                    }
                    relation {
                        id
                    }
                }
            }
            tags {
                edges {
                    node {
                        id
                        tag_type
                        value
                        color
                    }
                    relation {
                        id
                    }
                }
            }
            externalReferences {
                edges {
                    node {
                        id
                        entity_type
                        stix_id_key
                        source_name
                        description
                        url
                        hash
                        external_id
                        created
                        modified
                    }
                    relation {
                        id
                    }
                }
            }        
        """

    """
        List Attack-Pattern objects

        :param filters: the filters to apply
        :param search: the search keyword
        :param first: return the first n rows from the after ID (or the beginning if not set)
        :param after: ID of the first row for pagination
        :return List of Attack-Pattern objects
    """

    def list(self, **kwargs):
        filters = kwargs.get("filters", None)
        search = kwargs.get("search", None)
        first = kwargs.get("first", 500)
        after = kwargs.get("after", None)
        order_by = kwargs.get("orderBy", None)
        order_mode = kwargs.get("orderMode", None)
        get_all = kwargs.get("getAll", False)
        if get_all:
            first = 500

        self.opencti.log(
            "info", "Listing Attack-Patterns with filters " + json.dumps(filters) + "."
        )
        query = (
            """
            query AttackPatterns($filters: [AttackPatternsFiltering], $search: String, $first: Int, $after: ID, $orderBy: AttackPatternsOrdering, $orderMode: OrderingMode) {
                attackPatterns(filters: $filters, search: $search, first: $first, after: $after, orderBy: $orderBy, orderMode: $orderMode) {
                    edges {
                        node {
                            """
            + self.properties
            + """
                        }
                    }
                    pageInfo {
                        startCursor
                        endCursor
                        hasNextPage
                        hasPreviousPage
                        globalCount
                    }
                }
            }
        """
        )
        result = self.opencti.query(
            query,
            {
                "filters": filters,
                "search": search,
                "first": first,
                "after": after,
                "orderBy": order_by,
                "orderMode": order_mode,
            },
        )
        if get_all:
            final_data = []
            data = self.opencti.process_multiple(result["data"]["attackPatterns"])
            final_data = final_data + data
            while result["data"]["attackPatterns"]["pageInfo"]["hasNextPage"]:
                after = result["data"]["attackPatterns"]["pageInfo"]["endCursor"]
                self.opencti.log("info", "Listing Attack-Patterns after " + after)
                result = self.opencti.query(
                    query,
                    {
                        "filters": filters,
                        "search": search,
                        "first": first,
                        "after": after,
                        "orderBy": order_by,
                        "orderMode": order_mode,
                    },
                )
                data = self.opencti.process_multiple(result["data"]["attackPatterns"])
                final_data = final_data + data
            return final_data
        else:
            return self.opencti.process_multiple(result["data"]["attackPatterns"])

    """
        Read a Attack-Pattern object
        
        :param id: the id of the Attack-Pattern
        :param filters: the filters to apply if no id provided
        :return Attack-Pattern object
    """

    def read(self, **kwargs):
        id = kwargs.get("id", None)
        filters = kwargs.get("filters", None)
        custom_attributes = kwargs.get("customAttributes", None)
        if id is not None:
            self.opencti.log("info", "Reading Attack-Pattern {" + id + "}.")
            query = (
                """
                query AttackPattern($id: String!) {
                    attackPattern(id: $id) {
                        """
                + (
                    custom_attributes
                    if custom_attributes is not None
                    else self.properties
                )
                + """
                    }
                }
             """
            )
            result = self.opencti.query(query, {"id": id})
            return self.opencti.process_multiple_fields(result["data"]["attackPattern"])
        elif filters is not None:
            result = self.list(filters=filters)
            if len(result) > 0:
                return result[0]
            else:
                return None
        else:
            self.opencti.log(
                "error", "[opencti_attack_pattern] Missing parameters: id or filters"
            )
            return None

    """
        Create a Attack-Pattern object

        :param name: the name of the Attack Pattern
        :return Attack-Pattern object
    """

    def create_raw(self, **kwargs):
        name = kwargs.get("name", None)
        description = kwargs.get("description", None)
        alias = kwargs.get("alias", None)
        platform = kwargs.get("platform", None)
        required_permission = kwargs.get("required_permission", None)
        external_id = kwargs.get("external_id", None)
        id = kwargs.get("id", None)
        stix_id_key = kwargs.get("stix_id_key", None)
        created = kwargs.get("created", None)
        modified = kwargs.get("modified", None)
        created_by_ref = kwargs.get("createdByRef", None)
        marking_definitions = kwargs.get("markingDefinitions", None)
        kill_chain_phases = kwargs.get("killChainPhases", None)

        if name is not None and description is not None:
            self.opencti.log("info", "Creating Attack-Pattern {" + name + "}.")
            query = (
                """
                mutation AttackPatternAdd($input: AttackPatternAddInput) {
                    attackPatternAdd(input: $input) {
                        """
                + self.properties
                + """
                    }
                }
            """
            )
            result = self.opencti.query(
                query,
                {
                    "input": {
                        "name": name,
                        "description": description,
                        "alias": alias,
                        "platform": platform,
                        "required_permission": required_permission,
                        "external_id": external_id,
                        "internal_id_key": id,
                        "stix_id_key": stix_id_key,
                        "created": created,
                        "modified": modified,
                        "createdByRef": created_by_ref,
                        "markingDefinitions": marking_definitions,
                        "killChainPhases": kill_chain_phases,
                    }
                },
            )
            return self.opencti.process_multiple_fields(
                result["data"]["attackPatternAdd"]
            )
        else:
            self.opencti.log(
                "error",
                "[opencti_attack_pattern] Missing parameters: name and description",
            )

    """
        Create a Attack-Pattern object only if it not exists, update it on request

        :param name: the name of the Attack-Pattern
        :return Attack-Pattern object
    """

    def create(self, **kwargs):
        name = kwargs.get("name", None)
        description = kwargs.get("description", None)
        alias = kwargs.get("alias", None)
        platform = kwargs.get("platform", None)
        required_permission = kwargs.get("required_permission", None)
        external_id = kwargs.get("external_id", None)
        id = kwargs.get("id", None)
        stix_id_key = kwargs.get("stix_id_key", None)
        created = kwargs.get("created", None)
        modified = kwargs.get("modified", None)
        created_by_ref = kwargs.get("createdByRef", None)
        marking_definitions = kwargs.get("markingDefinitions", None)
        kill_chain_phases = kwargs.get("killChainPhases", None)
        update = kwargs.get("update", False)
        custom_attributes = """
            id
            entity_type
            name
            description 
            alias
            ... on AttackPattern {
                platform
                required_permission
                external_id
                
            }
        """
        object_result = self.opencti.stix_domain_entity.get_by_stix_id_or_name(
            types=["Attack-Pattern"],
            stix_id_key=stix_id_key,
            name=name,
            customAttributes=custom_attributes,
        )
        if object_result is None and external_id is not None:
            object_result = self.read(
                filters=[{"key": "external_id", "values": [external_id]}]
            )
        if object_result is not None:
            if update:
                # name
                if object_result["name"] != name:
                    self.opencti.stix_domain_entity.update_field(
                        id=object_result["id"], key="name", value=name
                    )
                    object_result["name"] = name
                # description
                if object_result["description"] != description:
                    self.opencti.stix_domain_entity.update_field(
                        id=object_result["id"], key="description", value=description
                    )
                    object_result["description"] = description
                # alias
                if alias is not None and object_result["alias"] != alias:
                    if "alias" in object_result:
                        new_aliases = object_result["alias"] + list(
                            set(alias) - set(object_result["alias"])
                        )
                    else:
                        new_aliases = alias
                    self.opencti.stix_domain_entity.update_field(
                        id=object_result["id"], key="alias", value=new_aliases
                    )
                    object_result["alias"] = new_aliases
                # platform
                if platform is not None and object_result["platform"] != platform:
                    self.opencti.stix_domain_entity.update_field(
                        id=object_result["id"], key="platform", value=platform
                    )
                    object_result["platform"] = platform
                # required_permission
                if (
                    required_permission is not None
                    and object_result["required_permission"] != required_permission
                ):
                    self.opencti.stix_domain_entity.update_field(
                        id=object_result["id"],
                        key="required_permission",
                        value=required_permission,
                    )
                    object_result["required_permission"] = required_permission
                # external_id
                if (
                    external_id is not None
                    and object_result["external_id"] != external_id
                ):
                    self.opencti.stix_domain_entity.update_field(
                        id=object_result["id"], key="external_id", value=external_id
                    )
                    object_result["external_id"] = external_id
            return object_result
        else:
            return self.create_raw(
                name=name,
                description=description,
                alias=alias,
                platform=platform,
                required_permission=required_permission,
                external_id=external_id,
                id=id,
                stix_id_key=stix_id_key,
                created=created,
                modified=modified,
                createdByRef=created_by_ref,
                markingDefinitions=marking_definitions,
                killChainPhases=kill_chain_phases,
            )

    """
        Import an Attack-Pattern object from a STIX2 object

        :param stixObject: the Stix-Object Attack-Pattern
        :return Attack-Pattern object
    """

    def import_from_stix2(self, **kwargs):
        stix_object = kwargs.get("stixObject", None)
        extras = kwargs.get("extras", {})
        update = kwargs.get("update", False)
        if stix_object is not None:
            # Extract external ID
            external_id = None
            if CustomProperties.EXTERNAL_ID in stix_object:
                external_id = stix_object[CustomProperties.EXTERNAL_ID]
            if "external_references" in stix_object:
                for external_reference in stix_object["external_references"]:
                    if (
                        external_reference["source_name"] == "mitre-attack"
                        or external_reference["source_name"] == "mitre-pre-attack"
                        or external_reference["source_name"] == "amitt-attack"
                    ):
                        external_id = external_reference["external_id"]
            return self.create(
                name=stix_object["name"],
                description=self.opencti.stix2.convert_markdown(
                    stix_object["description"]
                )
                if "description" in stix_object
                else "",
                alias=self.opencti.stix2.pick_aliases(stix_object),
                platform=stix_object["x_mitre_platforms"]
                if "x_mitre_platforms" in stix_object
                else stix_object["x_amitt_platforms"]
                if "x_amitt_platforms" in stix_object
                else None,
                required_permission=stix_object["x_mitre_permissions_required"]
                if "x_mitre_permissions_required" in stix_object
                else None,
                external_id=external_id,
                id=stix_object[CustomProperties.ID]
                if CustomProperties.ID in stix_object
                else None,
                stix_id_key=stix_object["id"] if "id" in stix_object else None,
                created=stix_object["created"] if "created" in stix_object else None,
                modified=stix_object["modified"] if "modified" in stix_object else None,
                createdByRef=extras["created_by_ref_id"]
                if "created_by_ref_id" in extras
                else None,
                markingDefinitions=extras["marking_definitions_ids"]
                if "marking_definitions_ids" in extras
                else [],
                killChainPhases=extras["kill_chain_phases_ids"]
                if "kill_chain_phases_ids" in extras
                else [],
                update=update,
            )
        else:
            self.opencti.log(
                "error", "[opencti_attack_pattern] Missing parameters: stixObject"
            )

    """
        Export an Attack-Pattern object in STIX2
    
        :param id: the id of the Attack-Pattern
        :return Attack-Pattern object
    """

    def to_stix2(self, **kwargs):
        id = kwargs.get("id", None)
        mode = kwargs.get("mode", "simple")
        max_marking_definition_entity = kwargs.get(
            "max_marking_definition_entity", None
        )
        entity = kwargs.get("entity", None)
        if id is not None and entity is None:
            entity = self.read(id=id)
        if entity is not None:
            attack_pattern = dict()
            attack_pattern["id"] = entity["stix_id_key"]
            attack_pattern["type"] = "attack-pattern"
            attack_pattern["spec_version"] = SPEC_VERSION
            if self.opencti.not_empty(entity["external_id"]):
                attack_pattern[CustomProperties.EXTERNAL_ID] = entity["external_id"]
            attack_pattern["name"] = entity["name"]
            if self.opencti.not_empty(entity["stix_label"]):
                attack_pattern["labels"] = entity["stix_label"]
            else:
                attack_pattern["labels"] = ["attack-pattern"]
            if self.opencti.not_empty(entity["description"]):
                attack_pattern["description"] = entity["description"]
            attack_pattern["created"] = self.opencti.stix2.format_date(
                entity["created"]
            )
            attack_pattern["modified"] = self.opencti.stix2.format_date(
                entity["modified"]
            )
            if self.opencti.not_empty(entity["platform"]):
                attack_pattern["x_mitre_platforms"] = entity["platform"]
            if self.opencti.not_empty(entity["required_permission"]):
                attack_pattern["x_mitre_permissions_required"] = entity[
                    "required_permission"
                ]
            if self.opencti.not_empty(entity["alias"]):
                attack_pattern[CustomProperties.ALIASES] = entity["alias"]
            attack_pattern[CustomProperties.ID] = entity["id"]
            return self.opencti.stix2.prepare_export(
                entity, attack_pattern, mode, max_marking_definition_entity
            )
        else:
            self.opencti.log(
                "error", "[opencti_attack_pattern] Missing parameters: id or entity"
            )
